import { Renderer } from "@coreserver/utils";
import * as path from "path";
import * as fs from "fs";
export class VueRenderer extends Renderer {
    renderer;
    readyPromise;
    step:number=0;
    constructor(config){
        super({
            default : { 
                client : "./build/webpack.client.config.js",
                server : "./build/webpack.server.config.js",
                bundle : "./dist/vue-ssr-server-bundle.json",
                manifest : "./dist/vue-ssr-client-manifest.json",
                template :  "./src/index.template.html",
                renderer : {
                    cache: require("lru-cache")({
                        max: 1000,
                        maxAge: 1000 * 60 * 15
                    }),
                    basedir: "./dist",
                    runInNewContext: false
                }
             } ,
            userDefined : config
        })
        
    }
    SetupDevServer(app, cb) {
        let bundle
        let clientManifest
    
        const webpack = require("webpack")
        const MFS = require("memory-fs")
        const clientConfig = require(this.pluginConfig.client)
        const serverConfig = require(this.pluginConfig.server)

        let resolve
        let resolved = false
        const readyPromise = new Promise((r) => {
            resolve = r
        })
        const ready = (...args) => {
            if (!resolved) resolve()
            cb(...args)
        }
    
        // modify client config to work with hot middleware
        clientConfig.entry.app = ["webpack-hot-middleware/client", clientConfig.entry.app]
        clientConfig.output.filename = "[name].js"
        clientConfig.plugins.push(
            new webpack.HotModuleReplacementPlugin(),
            new webpack.NoEmitOnErrorsPlugin()
        )
 
        // dev middleware
        const clientCompiler = webpack(clientConfig)
        const devMiddleware = require("webpack-dev-middleware")(clientCompiler, {
            publicPath: clientConfig.output.publicPath,
            noInfo: true
        })
        app.use(devMiddleware)
    
    
        clientCompiler.plugin("done", () => {
            const fs = devMiddleware.fileSystem
            const readFile = (file) => mfs.readFileSync(path.join(clientConfig.output.path, file), "utf-8")
            clientManifest = JSON.parse(readFile("vue-ssr-client-manifest.json"))
            if (bundle) {
                ready(bundle, {
                    clientManifest
                })
            }
        })
    
        // hot middleware
        app.use(require("webpack-hot-middleware")(clientCompiler))
    
        // watch and update server renderer
        const serverCompiler = webpack(serverConfig)
        const mfs = new MFS()
        serverCompiler.outputFileSystem = mfs
        serverCompiler.watch({}, (err, stats) => {
            if (err) throw err
            stats = stats.toJson()
            stats.errors.forEach((err) => console.error(err))
            stats.warnings.forEach((err) => console.warn(err))
            const readFile = (file) => mfs.readFileSync(path.join(clientConfig.output.path, file), "utf-8")
            // read bundle generated by vue-ssr-webpack-plugin
                bundle = JSON.parse(readFile("vue-ssr-server-bundle.json"))
            if (clientManifest) {
                cb(bundle, { clientManifest })
            }
        })
    
        return readyPromise
    }
    
    Init(config){
        super.Init(config);
        var keys = ["client" , "server" , "index" ,"bundle" , "manifest" , "basedir"];
        keys.forEach(k => {
            if(this.pluginConfig[k] && !path.isAbsolute(this.pluginConfig[k]))
                this.pluginConfig[k] = this.GetPathFromRoot(this.pluginConfig[k]);
        })
        this.pluginConfig.renderer.template = this.pluginConfig.template;
        if(this.pluginConfig.basedir && !path.isAbsolute(this.pluginConfig.basedir))
            this.pluginConfig.basedir = this.GetPathFromRoot(this.pluginConfig.basedir);
    }
    CreateRenderer(_bundle = null, _options = null){
        const bundle = _bundle || require(this.pluginConfig.bundle)
       
        var options = _options ;
        if(!options)
        {
            const clientManifest = require(this.pluginConfig.manifest)
            options =   Object.assign({ clientManifest }, this.pluginConfig.renderer);
        }        
        this.renderer = require("vue-server-renderer").createBundleRenderer(bundle, options);
    }
    Start(app) {
        super.Start(app);
        if(this.Config.env.DEVELOPPMENT)
        {
            this.readyPromise = this.SetupDevServer(app, (bundle, options) => {
                fs.readFile(this.pluginConfig.template, "utf-8", (err, data) => {
                    this.pluginConfig.renderer.template = data;
                    this.step++; 
                    this.CreateRenderer(bundle, options)
                });
               
            })
        }
        else{
            if(fs.existsSync(this.pluginConfig.template))
            fs.readFile(this.pluginConfig.template, "utf-8", (err, data) => {
                this.pluginConfig.renderer.template = data;
                this.step++;
                this.CreateRenderer()
                this.readyPromise= Promise.resolve();
            });
            else
                throw Error("Template path not found ," + this.pluginConfig.template)
        }
       
            
       

       
    }
    async Handle( req, res , next){
        if(this.step == 0)
            return res.send("App is building...");
        
        await this.readyPromise;
        
        var context = {
            url: req.url
        };
      
        this.renderer.renderToString(context, (err, html) => {
            if (err) 
                return next(err);

            res.setHeader("Content-Type", "text/html");
            res.status(200)
            res.end(html)
        })
    }
}